# Copyright (c) 2012 Nick Stenning, http://whiteink.com/
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# Author: Nick Stenning
# Site: https://github.com/nickstenning/honcho
#
# Modifications by Silas Sewell

from __future__ import print_function

import signal
import subprocess
import sys
from Queue import Queue, Empty
from threading import Thread


class Printer(object):
    def __init__(self, output=sys.stdout, name='unknown', width=0):
        self.output = output
        self.name = name
        self.width = width

        self._write_prefix = True

    def write(self, *args, **kwargs):
        new_args = []

        for arg in args:
            lines = arg.split('\n')
            lines = [self._prefix() + l if l else l for l in lines]
            new_args.append('\n'.join(lines))

        self.output.write(*new_args, **kwargs)

    def _prefix(self):
        return '{name} | '.format(name=self.name.ljust(self.width))


class Process(subprocess.Popen):
    """

    A simple utility wrapper around subprocess.Popen that stores
    a number of attributes needed by Honcho.

    """
    def __init__(self, cmd, name=None, *args, **kwargs):
        self.name = name
        self.reader = None
        self.printer = None

        defaults = {
            'stdout': subprocess.PIPE,
            'stderr': subprocess.STDOUT,
            'shell': True,
            'bufsize': 1,
            'close_fds': True
        }
        defaults.update(kwargs)

        super(Process, self).__init__(cmd, *args, **defaults)


class ProcessManager(object):
    """

    Here's where the business happens. The ProcessManager multiplexes and
    pretty-prints the output from a number of Process objects, typically added
    using the add_process() method.

    Example:

        pm = ProcessManager()
        pm.add_process('name', 'ruby server.rb')
        pm.add_process('name', 'python worker.py')

        pm.loop()

    """
    def __init__(self):
        self.processes = []
        self.queue = Queue()
        self.system_printer = Printer(sys.stdout, name='system')

        self._terminating = False

    def add_process(self, name, cmd):
        """

        Add a process to this manager instance:

        Arguments:

        name        - a human-readable identifier for the process
                      (e.g. 'worker'/'server')
        cmd         - the command-line used to run the process
                      (e.g. 'python run.py')

        """
        self.processes.append(Process(cmd, name))

    def loop(self):
        """

        Enter the main loop of the program. This will print the multiplexed
        output of all the processes in this ProcessManager to sys.stdout, and
        will block until all the processes have completed.

        If one process terminates, all the others will be terminated by
        Honcho, and loop() will return.

        """

        self._init_readers()
        self._init_printers()

        for proc in self.processes:
            print("started with pid {0:d}".format(proc.pid), file=proc.printer)

        while self._process_count() > 0:
            try:
                proc, line = self.queue.get(timeout=0.1)
            except Empty:
                pass
            except KeyboardInterrupt:
                print("SIGINT received", file=sys.stderr)
                self.terminate()
            else:
                print(line, end='', file=proc.printer)

                if proc.poll() is not None:
                    print('process terminated', file=proc.printer)
                    self.terminate()

    def terminate(self):
        """

        Terminate all the child processes of this ProcessManager, bringing the
        loop() to an end.

        """
        if self._terminating:
            return False

        self._terminating = True

        print("sending SIGTERM to all processes", file=self.system_printer)
        for proc in self.processes:
            if proc.poll() is None:
                print("sending SIGTERM to pid {0:d}".format(proc.pid),
                    file=self.system_printer)
                proc.terminate()

        def kill(signum, frame):
            # If anything is still alive, SIGKILL it
            for proc in self.processes:
                if proc.poll() is None:
                    print("sending SIGKILL to pid {0:d}".format(proc.pid),
                        file=self.system_printer)
                    proc.kill()

        signal.signal(signal.SIGALRM, kill)
        signal.alarm(5)

    def _process_count(self):
        return [p.poll() for p in self.processes].count(None)

    def _init_readers(self):
        for proc in self.processes:
            t = Thread(target=_enqueue_output, args=(proc, self.queue))
            t.daemon = True  # thread dies with the program
            t.start()

    def _init_printers(self):
        width = max(len(p.name) for p in self.processes)
        width = max(width, len(self.system_printer.name))

        self.system_printer.width = width

        for proc in self.processes:
            proc.printer = Printer(sys.stdout,
                                   name=proc.name,
                                   width=width)


def _enqueue_output(proc, queue):
    for line in iter(proc.stdout.readline, b''):
        if not line.endswith('\n'):
            line += '\n'
        queue.put((proc, line))
    proc.stdout.close()



import os
import unittest
from rock.config import Config

tests_path = os.path.join(os.path.dirname(__file__))

env_path = os.path.join(tests_path, 'assets', 'env')
project_path = os.path.join(tests_path, 'assets', 'project')
data_path = os.path.join(tests_path, 'assets', 'data')


class ConfigTestCase(unittest.TestCase):

    def setup_test(self):
        Config.MOUNT = os.path.join(env_path, 'test')
        Config.DATA = os.path.join(data_path, 'test')
        self.path = os.path.join(project_path, 'simple')
        return Config({'path': self.path})

    def test_values(self):
        c = self.setup_test()
        self.assertEqual(c['path'], self.path)
        self.assertEqual(c['runtime'], 'test123')
        self.assertEqual(c['runtime_type'], 'test')
        # env
        env = c['env']
        self.assertEqual(env['PATH'],
            '%s/bindir:/opt/rock/runtime/test123/usr/bin:${PATH}' %
            self.path)
        self.assertEqual(env['BUILD_PATH'], 'build_path')
        self.assertEqual(env['TEST_PATH'], 'test_path')
        self.assertEqual(env['PROJECT_PATH'], self.path)
        # tasks
        self.assertEqual(c['build'].strip(), 'build')
        self.assertEqual(c['build_deployment'].strip(), 'build deployment')
        self.assertEqual(c['clean'].strip(), 'clean')
        self.assertEqual(c['test'].strip(), 'test')



import collections
import copy
import os
import string
import yaml
from rock.exceptions import ConfigError


class Config(collections.Mapping):

    MOUNT = '/'
    DATA = os.path.join(os.path.dirname(__file__), 'data')

    def __init__(self, data):
        self._data = data
        self._setup = False

    def parse(self, path, require_exists=True, require_parses=True):
        if not os.path.isfile(path):
            if require_exists:
                raise ConfigError('not found: ' + path)
            else:
                return None
        try:
            with open(path) as f:
                return yaml.load(f)
        except Exception, error:
            if require_parses:
                raise ConfigError('parse error: ' + path)

    def merge(self, src, dst):
        if src is None:
            return dst
        if 'env' in src:
            # ensure env is a dict of strings
            if not (isinstance(src['env'], dict) or
                    all(map(lambda v: isinstance(v, basestring),
                    src['env'].values()))):
                raise ConfigError('env must be an associative array of ' +
                    'strings')
            if 'env' not in dst:
                dst['env'] = {}
            # evaluate env variables
            for name, value in src['env'].items():
                dst['env'][name] = string.Template(
                    src['env'][name]).safe_substitute(**dst['env'])
            del src['env']
        dst.update(src)

    def setup(self):
        if self._setup: return
        self._setup = True
        # setup configuration
        data = {}
        # runtime
        yml_path = ('path' in self._data and
            os.path.join(self._data['path'], '.rock.yml'))
        if yml_path and os.path.isfile(yml_path):
            data = self.parse(yml_path)
            data.update(self._data)
        else:
            data = copy.deepcopy(self._data)
        if 'runtime' in data and 'runtime_type' not in data:
            data['runtime_type'] = data['runtime'].rstrip('0123456789')
        # project
        for name in ('path', 'runtime', 'runtime_type'):
            if name not in data:
                raise ConfigError('%s is required' % name)
        # paths
        runtime_path = os.path.join(self.MOUNT, 'opt', 'rock', 'runtime',
            data['runtime'])
        etc_path = os.path.join(self.MOUNT, 'etc', 'rock', 'runtime')
        runtime_type_yml = data['runtime_type'] + '.yml'
        runtime_yml = data['runtime'] + '.yml'
        # ensure runtime exists
        if not os.path.isdir(runtime_path):
            raise ConfigError("runtime path doesn't exist")
        # configs
        runtime_config = self.parse(os.path.join(runtime_path,
            'rock.yml'))
        rock_type_config = self.parse(os.path.join(self.DATA, 'runtime',
            runtime_type_yml), require_exists=False)
        rock_config = self.parse(os.path.join(self.DATA, 'runtime',
            runtime_yml), require_exists=False)
        etc_type_config = self.parse(os.path.join(etc_path,
            runtime_type_yml), require_exists=False)
        etc_config = self.parse(os.path.join(etc_path, runtime_yml),
            require_exists=False)
        # merge
        self._data = {
            'env': {
                'PROJECT_PATH': data['path'],
            },
        }
        # merge runtime
        self.merge(runtime_config, self._data)
        # merge runtime config
        if rock_config or etc_config:
            self.merge(rock_config, self._data)
            self.merge(etc_config, self._data)
        else:
            self.merge(rock_type_config, self._data)
            self.merge(etc_type_config, self._data)
        # merge project
        self.merge(data, self._data)

    def __contains__(self, *args, **kwargs):
        self.setup()
        return self._data.__contains__(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        self.setup()
        return self._data.__getitem__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        self.setup()
        return self._data.__getitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        self.setup()
        return self._data.__len__(*args, **kwargs)



import os
import pipes
import string
import sys
import yaml
from rock.config import Config
from rock.exceptions import ConfigError
from rock.process import ProcessManager
from rock.utils import Shell


class Project(object):

    def __init__(self, config):
        self.config = Config(config)

    def execute(self, command):
        with Shell() as s:
            # declare builtin functions
            s.write('warn() { echo "$@" >&2; }')
            s.write('die() { warn "$@"; exit 1; }')
            # print commands as they're run
            if self.config.get('verbose'):
                s.write('set -o verbose')
            # don't execute commands, just print them
            if self.config.get('dry_run'):
                s.write('set -o noexec')
            # exit with error if any one command fails
            s.write('set -o errexit')
            # setup environment variables
            if self.config.get('env'):
                # blank line before exports
                s.write('')
                for name, value in self.config['env'].items():
                    s.write('export %s="%s"' % (name, value))
            # blank line before command
            s.write('')
            # run command and wait for results
            if isinstance(command, (list, tuple)):
                s.write(' '.join(command))
            else:
                s.write(command)

    def execute_type(self, name, *args):
        section = '%s_%s' % (name, args[0]) if len(args) > 0 else name
        if section not in self.config:
            raise ConfigError('section not found: %s' % section)
        self.execute(self.config[section])

    def execute_section(self, name):
        section = self.config.get(name)

        if isinstance(section, dict):
            pm = ProcessManager()
            for name, value in section.items():
                pm.add_process(name, '%s run %s' %
                    (sys.argv[0], pipes.quote(value)))
            pm.loop()
        else:
            self.execute(section)

    def build(self, *args):
        self.execute_type('build', *args)

    def clean(self, *args):
        self.execute_type('clean', *args)

    def run(self, args):
        if len(args) == 0 and 'run' in self.config:
            self.execute_section('run')
        elif len(args) == 1 and 'run_%s' % args[0] in self.config:
            self.execute_section('run_%s' % args[0])
        else:
            self.execute(args)

    def test(self, *args):
        self.execute_type('test', *args)



import StringIO
import os


class Shell(object):

    def __enter__(self):
        self.stdin = StringIO.StringIO()
        return self

    def __exit__(self, type, value, traceback):
        os.execl('/usr/bin/env', 'bash', 'bash', '-c', self.stdin.getvalue())

    def write(self, text):
        self.stdin.write(text + '\n')
